#
# Copyright (c) 2020 Xilinx, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#


# Set default logging handler to avoid "No handler found" warnings.
import logging
from logging import NullHandler

logging.getLogger(__name__).addHandler(NullHandler())

# Setup namespace package
from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)  # type: ignore]

from ._version import __version__
